@@ -29,7 +29,7 @@
 #include "dvdauthor.h"
 #include "da-internal.h"
 
-static const char RCSID[]="$Id: //depot/dvdauthor/src/dvdvob.c#49 $";
+static const char RCSID[]="$Id: //depot/dvdauthor/src/dvdvob.c#55 $";
 
 
 struct colorremap {
@@ -43,8 +43,8 @@ struct vscani {
     int firstgop,firsttemporal,lastadjust,adjustfields;
 };
 
-static int timeline[19]={1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
-                         20,60,120,240};
+static pts_t timeline[19]={1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
+                           20,60,120,240};
 
 #define BIGWRITEBUFLEN (16*2048)
 static unsigned char bigwritebuf[BIGWRITEBUFLEN];
@@ -53,9 +53,23 @@ static int writefile=-1;
 
 static unsigned char videoslidebuf[15]={255,255,255,255, 255,255,255, 0,0,0,0, 0,0,0,0};
 
-static int calcpts(struct vobgroup *va,int basepts,int nfields)
+static pts_t calcpts(struct vobgroup *va,int cancomplain,int *didcomplain,pts_t *align,pts_t basepts,int nfields)
 {
-    return basepts+(nfields*getframepts(va))/2;
+    // I assume pts should round down?  That seems to be how mplex deals with it
+    // also see later comment
+
+    int fpts=getframepts(va);
+    int bpframe=(basepts*2-*align+fpts/2)/fpts;
+    if( (*align+bpframe*fpts)/2 != basepts ) {
+        if( !*didcomplain ) {
+            if( cancomplain )
+                fprintf(stderr,"WARN: Video PTS does not line up on a multiple of a field.\n");
+            *didcomplain=1;
+        }
+        *align=basepts;
+    } else
+        nfields += bpframe;
+    return (*align+nfields*fpts)/2;
 }
 
 static int findnextvideo(struct vob *va, int cur, int dir)
@@ -77,7 +91,7 @@ static int findnextvideo(struct vob *va, int cur, int dir)
     }
 }
 
-static int findaudsect(struct vob *va,int aind,int pts0,int pts1)
+static int findaudsect(struct vob *va,int aind,pts_t pts0,pts_t pts1)
 {
     struct audchannel *ach=&va->audch[aind];
     int l=0,h=ach->numaudpts-1;
@@ -114,7 +128,7 @@ static int findvobubysect(struct vob *va,int sect)
     return l;
 }
 
-static int findspuidx(struct vob *va,int ach,int pts0)
+static int findspuidx(struct vob *va,int ach,pts_t pts0)
 {
     int l=0,h=va->audch[ach].numaudpts-1;
 
@@ -159,40 +173,40 @@ static unsigned int getsect(struct vob *va,int curvobnum,int jumpvobnum,int skip
         |skip;
 }
 
-static int readscr(const unsigned char *buf)
+static pts_t readscr(const unsigned char *buf)
 {
-    return ((buf[0]&0x18)<<27)|
-            ((buf[0]&3)<<28)|
-            (buf[1]<<20)|
-            ((buf[2]&0xf8)<<12)|
-            ((buf[2]&3)<<13)|
-            (buf[3]<<5)|
-            ((buf[4]&0xf8)>>3);
+    return (((pts_t)(buf[0]&0x38))<<27)|
+        ((buf[0]&3)<<28)|
+        (buf[1]<<20)|
+        ((buf[2]&0xf8)<<12)|
+        ((buf[2]&3)<<13)|
+        (buf[3]<<5)|
+        ((buf[4]&0xf8)>>3);
 }
 
-static void writescr(unsigned char *buf,unsigned int scr)
+static void writescr(unsigned char *buf,pts_t scr)
 {
-    buf[0]=((scr>>27)&0x18)|((scr>>28)&3)|68;
+    buf[0]=((scr>>27)&0x38)|((scr>>28)&3)|68;
     buf[1]=scr>>20;
     buf[2]=((scr>>12)&0xf8)|((scr>>13)&3)|4;
     buf[3]=scr>>5;
     buf[4]=((scr<<3)&0xf8)|(buf[4]&7);
 }
 
-static int readpts(const unsigned char *buf)
+static pts_t readpts(const unsigned char *buf)
 {
-    int a1,a2,a3,pts;
+    int a1,a2,a3;
     a1=(buf[0]&0xe)>>1;
     a2=((buf[1]<<8)|buf[2])>>1;
     a3=((buf[3]<<8)|buf[4])>>1;
-    pts=(((int64_t)a1)<<30)|
+
+    return (((pts_t)a1)<<30)|
         (a2<<15)|
         a3;
-    return pts;
 }
 
 
-static void writepts(unsigned char *buf,unsigned int pts)
+static void writepts(unsigned char *buf,pts_t pts)
 {
     buf[0]=((pts>>29)&0xe)|(buf[0]&0xf1); // this preserves the PTS / DTS / PTSwDTS top bits
     write2(buf+1,(pts>>14)|1);
@@ -211,7 +225,7 @@ static int findbutton(struct pgc *pg,char *dest,int dflt)
     return dflt;
 }
 
-static void transpose_ts(unsigned char *buf,int tsoffs)
+static void transpose_ts(unsigned char *buf,pts_t tsoffs)
 {
     // pack scr
     if( buf[0] == 0 &&
@@ -545,9 +559,9 @@ static void finishvideoscan(struct vobgroup *va,int vob,int prevsect,struct vsca
     closelastref(lastvi,vsi,prevsect);
 }
 
-static void printpts(int pts)
+static void printpts(pts_t pts)
 {
-    fprintf(stderr,"%d.%03d",pts/90000,(pts/90)%1000);
+    fprintf(stderr,"%d.%03d",(int)(pts/90000),(int)((pts/90)%1000));
 }
 
 static FILE *openvob(char *f,int *ispipe)
@@ -639,7 +653,7 @@ static void remapbyte(struct colorremap *cr,unsigned char *b)
     b[0]=remapcolor(cr,b[0]&15)|(remapcolor(cr,b[0]>>4)<<4);
 }
 
-static void procremap(struct colorremap *cr,unsigned char *b,int l,int *timespan)
+static void procremap(struct colorremap *cr,unsigned char *b,int l,pts_t *timespan)
 {
     while(l) {
         // fprintf(stderr,"INFO: state=%d, byte=%02x (%d)\n",cr->state,*b,*b);
@@ -807,7 +821,8 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
     crs=malloc(sizeof(struct colorremap)*32);
     for( vnum=0; vnum<va->numvobs; vnum++ ) {
         int i,j;
-        int hadfirstvobu=0,backoffs=0,lastscr=0;
+        int hadfirstvobu=0;
+        pts_t backoffs=0, lastscr=0;
         struct vob *s=va->vobs[vnum];
         int prevvidsect=-1;
         struct vscani vsi;
@@ -834,7 +849,14 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
                 fsect=-1;
             }
             buf=writegrabbuf();
-            if( 2048 != fread(buf,1,2048,h) ) {
+
+            i=fread(buf,1,2048,h);
+            if( i!=2048 ) {
+                if( i==-1 ) {
+                    fprintf(stderr,"\nERR:  Error while reading: %s\n",strerror(errno));
+                    exit(1);
+                } else if( i>0 )
+                    fprintf(stderr,"\nWARN: Partial sector read (%d bytes); discarding data.\n",i);
                 writeundo();
                 break;
             }
@@ -934,22 +956,20 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
                 writeundo();
                 continue;
             }
-
-            if( buf[0]==0 && buf[1]==0 && buf[2]==1 && buf[3]==0xba )
-            {
-              int newscr=readscr(buf+4);
-              if((newscr==0)&&(lastscr>0))
-              {
-                backoffs-=lastscr;
-                fprintf(stderr,"\nWARN:  SCR reset. New back offset = %d \n",backoffs);
-              }
-              else if( newscr < lastscr ) {
+            if( buf[0]==0 && buf[1]==0 && buf[2]==1 && buf[3]==0xba ) {
+                pts_t newscr=readscr(buf+4);
+                if((newscr==0)&&(lastscr>0))
+                {
+                  backoffs-=lastscr;
+                  fprintf(stderr,"\nWARN:  SCR reset. New back offset = %d \n",backoffs);
+                }
+                else if( newscr < lastscr ) {
 //                fprintf(stderr,"ERR:  SCR moves backwards, remultiplex input. %d<%d\n",newscr,lastscr);
 //                 exit(1);
-              }
-              lastscr=newscr;
-              if( !hadfirstvobu )
-                  backoffs=newscr;
+                }
+                lastscr=newscr;
+                if( !hadfirstvobu )
+                    backoffs=newscr;
             }
             transpose_ts(buf,-backoffs);
             if( fsect == -1 ) {
@@ -1057,7 +1077,7 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
                 buf[16] == 1 &&
                 ((buf[17] & 0xf8) == 0xc0 || buf[17]==0xbd) &&
                 buf[21] & 128 ) {
-                int pts0=readpts(buf+23),pts1,backpts1;
+                pts_t pts0=readpts(buf+23),pts1,backpts1;
                 int dptr=buf[22]+23,endop=read2(buf+18)+20;
                 int audch;
                 pts1=pts0;
@@ -1155,8 +1175,8 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
                         printpts(pts1);
                         fprintf(stderr,"\n");
                         ach->audpts[ach->numaudpts-1].pts[1]=pts0;
-                    noshow:
                     }
+                noshow:
                     ach->audpts[ach->numaudpts].pts[0]=pts0;
                     ach->audpts[ach->numaudpts].pts[1]=pts1;
                     ach->audpts[ach->numaudpts].sect=cursect;
@@ -1184,12 +1204,23 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
             fsect++;
         }
         switch(ispipe) {
-        case 0: fclose(h); break;
-        case 1: pclose(h); break;
+        case 0:
+            if (fclose(h)) {
+                fprintf(stderr,"\nERR:  Error reading from file: %s\n",strerror(errno));
+                exit(1);
+            }
+            break;
+        case 1:
+            if (pclose(h)) {
+                fprintf(stderr,"\nERR:  Error reading from pipe: %s\n",strerror(errno));
+                exit(1);
+            }
+            break;
         case 2: break;
         }
         if( s->numvobus ) {
-            int i,finalaudiopts;
+            int i;
+            pts_t finalaudiopts;
                 
             finishvideoscan(va,vnum,prevvidsect,&vsi);
             // find end of audio
@@ -1202,22 +1233,52 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
             }
                     
             // pin down all video vobus
-            for( i=0; i<s->numvobus; i++ ) {
-                struct vobuinfo *vi=s->vi+i;
-                if( vi->hasvideo ) {
-                    vi->videopts[0]=calcpts(va,vi->firstvideopts,-vi->firstIfield);
-                    vi->videopts[1]=calcpts(va,vi->videopts[0],vi->numfields);
+            // note: we make two passes; one assumes that the PTS for the
+            // first frame is exact; the other assumes that the PTS for
+            // the first frame is off by 1/2.  If both fail, then the third pass
+            // assumes things are exact and throws a warning
+            for( i=0; i<3; i++ ) {
+                pts_t pts_align=-1;
+                int complain=0, j;
+                
+                for( j=0; j<s->numvobus; j++ ) {
+                    struct vobuinfo *vi=s->vi+j;
+                    
+                    if( vi->hasvideo ) {
+                        if( pts_align==-1 ) {
+                            pts_align=vi->firstvideopts*2;
+                            if( i==1 ) {
+                                // I assume pts should round down?  That seems to be how mplex deals with it
+                                // also see earlier comment
 
-                    vi->sectpts[0]=vi->videopts[0];
-                    vi->sectpts[1]=vi->videopts[1];
+                                // since pts round down, then the alternative base we should try is
+                                // firstvideopts+0.5, thus increment
+                                pts_align++;
+                            }
+                            // MarkChapters will complain if firstIfield!=0
+                        }
+
+                        vi->videopts[0]=calcpts(va,i==2,&complain,&pts_align,vi->firstvideopts,-vi->firstIfield);
+                        vi->videopts[1]=calcpts(va,i==2,&complain,&pts_align,vi->firstvideopts,-vi->firstIfield+vi->numfields);
+                        // if this looks like a dud, abort and try the next pass
+                        if( complain && i<2 )
+                            break;
+
+                        vi->sectpts[0]=vi->videopts[0];
+                        vi->sectpts[1]=vi->videopts[1];
+                    }
                 }
+                if( !complain )
+                    break;
             }
             
             // guess at non-video vobus
             for( i=0; i<s->numvobus; i++ ) {
                 struct vobuinfo *vi=s->vi+i;
                 if( !vi->hasvideo ) {
-                    int j,k,firstaudiopts=-1;
+                    int j,k;
+                    pts_t firstaudiopts=-1,p;
+
                     for( j=0; j<32; j++ ) {
                         struct audchannel *ach=s->audch+j;
                         for( k=0; k<ach->numaudpts; k++ )
@@ -1232,33 +1293,33 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
                         firstaudiopts=readscr(vi->sectdata+4)+4*147; // 147 is roughly the minumum pts that must transpire between packets; we give a couple packets of buffer to allow the dvd player to process the data
                     }
                     if( i ) {
-                        j=firstaudiopts-s->vi[i-1].sectpts[0];
+                        pts_t frpts=getframepts(va);
+                        p=firstaudiopts-s->vi[i-1].sectpts[0];
                         // ensure this is a multiple of a framerate, just to be nice
-                        k=getframepts(va);
-                        j+=k-1;
-                        j-=j%k;
-                        j+=s->vi[i-1].sectpts[0];
-                        assert(j>=s->vi[i-1].sectpts[1]);
-                        s->vi[i-1].sectpts[1]=j;
+                        p+=frpts-1;
+                        p-=p%frpts;
+                        p+=s->vi[i-1].sectpts[0];
+                        assert(p>=s->vi[i-1].sectpts[1]);
+                        s->vi[i-1].sectpts[1]=p;
                     } else {
                         fprintf(stderr,"ERR:  Cannot infer pts for VOBU if there is no audio or video and it is the\nERR:  first VOBU.\n");
                         exit(1);
                     }
-                    vi->sectpts[0]=j;
+                    vi->sectpts[0]=p;
 
                     // if we can easily predict the end pts of this sector,
                     // then fill it in.  otherwise, let the next iteration do it
                     if( i+1==s->numvobus ) { // if this is the end of the vob, use the final audio pts as the last pts
                         if( finalaudiopts>vi->sectpts[0] )
-                            j=finalaudiopts;
+                            p=finalaudiopts;
                         else
-                            j=vi->sectpts[0]+getframepts(va); // add one frame of a buffer, so we don't have a zero (or less) length vobu
+                            p=vi->sectpts[0]+getframepts(va); // add one frame of a buffer, so we don't have a zero (or less) length vobu
                     } else if( s->vi[i+1].hasvideo ) // if the next vobu has video, use the start of the video as the end of this vobu
-                        j=s->vi[i+1].sectpts[0];
+                        p=s->vi[i+1].sectpts[0];
                     else // the next vobu is an audio only vobu, and will backfill the pts as necessary
                         continue;
-                    assert(j>vi->sectpts[0]);
-                    vi->sectpts[1]=j;
+                    assert(p>vi->sectpts[0]);
+                    vi->sectpts[1]=p;
                 }
             }
 
@@ -1292,7 +1353,14 @@ int FindVobus(char *fbase,struct vobgroup *va,int ismenu)
     return 1;
 }
 
-static int findnearestvobu(struct vobgroup *pg,struct vob *va,int pts)
+static pts_t pabs(pts_t pts)
+{
+    if( pts<0 )
+        return -pts;
+    return pts;
+}
+
+static int findnearestvobu(struct vobgroup *pg,struct vob *va,pts_t pts)
 {
     int l=0,h=va->numvobus-1,i;
 
@@ -1301,7 +1369,7 @@ static int findnearestvobu(struct vobgroup *pg,struct vob *va,int pts)
     pts+=va->vi[0].sectpts[0];
     i=findvobu(va,pts,l,h);
     if( i+1<va->numvobus && i>=0 &&
-        abs(pts-va->vi[i+1].sectpts[0])<abs(pts-va->vi[i].sectpts[0]) )
+        pabs(pts-va->vi[i+1].sectpts[0])<pabs(pts-va->vi[i].sectpts[0]) )
         i++;
     return i;
 }
@@ -1406,7 +1474,7 @@ void MarkChapters(struct vobgroup *va)
         }
 }
 
-static int getcellaudiopts(struct vobgroup *va,int vcid,int ach,int w)
+static pts_t getcellaudiopts(struct vobgroup *va,int vcid,int ach,int w)
 {
     struct vob *v=va->vobs[(vcid>>8)-1];
     struct audchannel *a=&v->audch[ach];
@@ -1428,7 +1496,7 @@ static int hasaudio(struct vobgroup *va,int vcid,int ach,int w)
     return a->numaudpts!=0;
 }
 
-static int getcellvideopts(struct vobgroup *va,int vcid,int w)
+static pts_t getcellvideopts(struct vobgroup *va,int vcid,int w)
 {
     struct vob *v=va->vobs[(vcid>>8)-1];
     int vi=0;
@@ -1442,7 +1510,7 @@ static int getcellvideopts(struct vobgroup *va,int vcid,int w)
     return v->vi[vi].sectpts[w];
 }
 
-static int calcaudiodiff(struct vobgroup *va,int vcid,int ach,int w)
+static pts_t calcaudiodiff(struct vobgroup *va,int vcid,int ach,int w)
 {
     return getcellvideopts(va,vcid,w)-getcellaudiopts(va,vcid,ach,w);
 }
@@ -1461,8 +1529,8 @@ int calcaudiogap(struct vobgroup *va,int vcid0,int vcid1,int ach)
             return 0;
 
         // we have problems if the second half has audio but the first doesn't
-        if( hasaudio(va,vcid0,ach,1) && !hasaudio(va,vcid1,ach,0) ) {
-            fprintf(stderr,"WARN: Transition from non-audio to audio VOB; assumjing discontinuity.\n");
+        if( !hasaudio(va,vcid0,ach,1) && hasaudio(va,vcid1,ach,0) ) {
+            fprintf(stderr,"WARN: Transition from non-audio to audio VOB; assuming discontinuity.\n");
             return 1;
         }
 
@@ -1477,7 +1545,8 @@ int calcaudiogap(struct vobgroup *va,int vcid0,int vcid1,int ach)
 void FixVobus(char *fbase,struct vobgroup *va,int ismenu)
 {
     int h=-1;
-    int i,j,pn,scr,fnum=-2;
+    int i,j,pn,fnum=-2;
+    pts_t scr;
     int vff,vrew,totvob,curvob;
     unsigned char *buf;
 
@@ -1576,7 +1645,9 @@ void FixVobus(char *fbase,struct vobgroup *va,int ismenu)
                 if( s>=0 ) {
                     s=s-vi->sector;
                     if( s > 0x1fff || s < -(0x1fff)) {
-                        fprintf(stderr,"\nWARN: audio sector out of range: %d (vobu #%d, pts %d)\n",s,i,vi->sectpts[0]);
+                        fprintf(stderr,"\nWARN: audio sector out of range: %d (vobu #%d, pts ",s,i);
+                        printpts(vi->sectpts[0]);
+                        fprintf(stderr,")\n");
                         s=0;
                     }
                     if( s < 0 )
